/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <guichaz@yahoo.fr>
 */

/*****************************
 * User configurable options *
 *****************************/

#include <string.h>             /* memcmp() */
#include <gdk/gdkkeysyms.h>     /* GDK_Escape, GDK_c, GDK_C, GDK_o, GDK_O */

#include "gliv.h"
#include "options.h"
#include "glade_options.h"
#include "gliv_image.h"
#include "gl_widget.h"
#include "history.h"
#include "windows.h"
#include "matrix.h"
#include "messages.h"
#include "rendering.h"
#include "scrollbars.h"
#include "next_image.h"
#include "rcfile.h"
#include "mnemonics.h"
#include "tree.h"

extern rt_struct *rt;
extern options_struct *options;
extern gliv_image *current_image;
extern GtkWidget *gl_widget;

/* Filled by the dialog. */
static options_struct *new_options;
static GtkWidget *options_dialog = NULL;

static gboolean save_options = FALSE;

static void toggle_delay(void)
{
    if (options->delay == 0)
        /* Enable. */
        gtk_widget_add_events(gl_widget, GDK_POINTER_MOTION_MASK);

    options->delay = new_options->delay;
    schedule_hide_cursor();
}

static void toggle_history(void)
{
    options->history_size = new_options->history_size;
    clean_history();
}

static void finish(gboolean apply)
{
    gint what = 0;
    gboolean bg_color_changed;

    gtk_widget_hide(options_dialog);

    if (apply) {

        if (options->fullscreen != new_options->fullscreen)
            toggle_fullscreen(new_options->fullscreen);

        if ((options->maximize == FALSE && new_options->maximize) ||
            (options->scaledown == FALSE && new_options->scaledown)) {

            options->maximize = new_options->maximize;
            options->scaledown = new_options->scaledown;

            matrix_set_max_zoom(-1, -1, TRUE);
            what |= REFRESH_IMAGE | REFRESH_STATUS | APPEND_HISTORY;
        }

        if (options->menu_bar != new_options->menu_bar)
            toggle_menu_bar();

        if (options->status_bar != new_options->status_bar)
            toggle_status_bar();

        if (options->scrollbars != new_options->scrollbars)
            toggle_scrollbars();

        if (options->dither != new_options->dither ||
            options->mipmap != new_options->mipmap) {

            options->dither = new_options->dither;
            options->mipmap = new_options->mipmap;

            reload_all_images();
            what |= REFRESH_IMAGE;
        }

        if (options->one_image == FALSE && new_options->one_image)
            unload_images();

        if (options->delay != new_options->delay)
            toggle_delay();

        if (options->history_size != new_options->history_size)
            toggle_history();

        if (options->duration != new_options->duration)
            if (slide_show_started()) {
                options->duration = new_options->duration;
                start_slide_show();
            }

        if (options->thumbnails != new_options->thumbnails ||
            options->thumb_width != new_options->thumb_width ||
            options->thumb_height != new_options->thumb_height)
            invalidate_last_tree();

        bg_color_changed = !gdk_color_equal(&options->bg_col,
                                            &new_options->bg_col);

        if (!gdk_color_equal(&options->alpha1, &new_options->alpha1) ||
            !gdk_color_equal(&options->alpha2, &new_options->alpha2)) {

            rt->alpha_checks_changed = TRUE;

            if (current_image != NULL && current_image->has_alpha)
                what |= REFRESH_IMAGE;
        }

        g_free(options);
        options = new_options;

        if (save_options)
            save_rc(options);

        if (bg_color_changed) {
            update_bg_color();
            what |= REFRESH_IMAGE;
        }
    } else
        g_free(new_options);

    gtk_widget_destroy(options_dialog);
    options_dialog = NULL;

    refresh(what);
}

static gboolean key_press_event(GtkWidget * unused, GdkEventKey * event)
{
    if (event->keyval == GDK_Escape) {
        finish(FALSE);
        return TRUE;
    }

    return FALSE;
}

/* Cancel and OK buttons */

static void add_cancel_button(GtkWidget * button)
{
    g_signal_connect_swapped(button, "clicked",
                             G_CALLBACK(finish), GINT_TO_POINTER(FALSE));
}

static void add_ok_button(GtkWidget * button)
{
    GTK_WIDGET_SET_FLAGS(GTK_WIDGET(button), GTK_CAN_DEFAULT);
    g_signal_connect_swapped(button, "clicked",
                             G_CALLBACK(finish), GINT_TO_POINTER(TRUE));

    g_signal_connect(button, "realize",
                     G_CALLBACK(gtk_widget_grab_default), NULL);
}

/* GtkColorSelection */

static void add_color_selection(GtkWidget * color_sel, GdkColor * col)
{
    gtk_color_selection_set_current_color(GTK_COLOR_SELECTION(color_sel), col);

    g_signal_connect(color_sel, "color-changed",
                     G_CALLBACK(gtk_color_selection_get_current_color), col);
}

/* GtkSpinButton */

static gboolean value_changed(GtkAdjustment * adj, gint * val)
{
    *val = (gint) gtk_adjustment_get_value(adj);
    return TRUE;
}

static void add_spin_button(GtkWidget * button, gint * value)
{
    GtkAdjustment *adj;

    adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(button));
    gtk_adjustment_set_value(adj, (gdouble) * value);

    g_signal_connect(adj, "value-changed", G_CALLBACK(value_changed), value);
}

/* GtkCheckButton */

static gboolean toggled_button(GtkToggleButton * button, gboolean * bool)
{
    *bool = gtk_toggle_button_get_active(button);
    return TRUE;
}

static void add_check_button(GtkWidget * button, gboolean * value)
{
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), *value);
    g_signal_connect(button, "toggled", G_CALLBACK(toggled_button), value);
}

#include "glade_options.c"

void show_options(void)
{
    if (options_dialog != NULL)
        return;

    new_options = g_memdup(options, sizeof(options_struct));
    options_dialog = create_options_dialog();

    g_signal_connect_swapped(options_dialog, "delete-event",
                             G_CALLBACK(finish), GINT_TO_POINTER(FALSE));

    g_signal_connect(options_dialog, "key-press-event",
                     G_CALLBACK(key_press_event), NULL);

    show_dialog(GTK_WIDGET(options_dialog), _("Options"));
}
